/*
 * *********************************************************************************************************************
 *
 *          !!
 *        .F88X
 *        X8888Y
 *      .}888888N;
 *        i888888N;        .:!              .I$WI:
 *          R888888I      .'N88~            i8}+8Y&8"l8i$8>8W~'>W8}8]KW+8IIN"8&
 *          .R888888I    .;N8888~          .X8'  "8I.!,/8"  !%NY8`"8I8~~8>,88I
 *            +888888N;  .8888888Y                                  "&&8Y.}8,
 *            ./888888N;  .R888888Y        .'}~    .>}'.`+>  i}!    "i'  +/'  .'i~  !11,.:">,  .~]!  .i}i
 *              ~888888%:  .I888888l      .]88~`1/iY88Ii+1'.R$8$8]"888888888>  Y8$  W8E  X8E  W8888'188Il}Y88$*
 *              18888888    E8888881    .]W%8$`R8X'&8%++N8i,8N%N8+l8%`  .}8N:.R$RE%N88N%N$K$R  188,FE$8%~Y88I
 *            .E888888I  .i8888888'      .:$8I;88+`E8R:/8N,.>881.`$8E/1/]N8X.Y8N`"KF&&FK!'88*."88K./$88%RN888+~
 *            8888888I  .,N888888~        ~88i"8W,!N8*.I88.}888%F,i$88"F88"  888:E8X.>88!i88>`888*.}Fl1]*}1YKi'
 *          i888888N'      I888Y          ]88;/EX*IFKFK88X  K8R  .l8W  88Y  ~88}'88E&%8W.X8N``]88!.$8K  .:W8I
 *        .i888888N;        I8Y          .&8$  .X88!  i881.:%888>I88  ;88]  +88+.';;;;:.Y88X  18N.,88l  .+88/
 *      .:R888888I
 *      .&888888I                                          Copyright (c) 2009-2020.  博锐尚格科技股份有限公司
 *        ~8888'
 *        .!88~                                                                     All rights reserved.
 *
 * *********************************************************************************************************************
 */

import { SPathCommand } from "./SPathCommand";
import { SPoint } from "./types/SPoint";

/**
 * Path 对象
 *
 * @author 庞利祥 <sybotan@126.com>
 */
export class SPath {
    /** 命令集合 */
    cmdList: SPathCommand[] = [];

    /**
     * 添加 path 命令
     *
     * @param cmd         命令
     * @param args        命令参数
     */
    private addCommand(cmd: string, ...args: number[]): void {
        this.cmdList.push({ command: cmd, args: args });
    }

    /**
     * 添加一条新路径到对当前路径。
     *
     * @param path  路径对象
     */
    addPath(path: SPath): void {
        this.cmdList = this.cmdList.concat(path.cmdList);
    }

    /**
     * 使笔点返回到当前子路径的起始点。它尝试从当前点到起始点绘制一条直线。 如果图形已经是封闭的或者只有一个点，那么此函数不会做任何操作。
     */
    closePath(): void {
        this.addCommand("Z");
    }

    /**
     * 移动当前点到指定位置
     *
     * @param x       X 坐标
     * @param y       Y 坐标
     */
    moveTo(x: number, y: number): void {
        this.addCommand("M", x, y);
    }

    /**
     * 从当前点向指定点画线
     *
     * @param x       X 坐标
     * @param y       Y 坐标
     */
    lineTo(x: number, y: number): void {
        this.addCommand("L", x, y);
    }

    /**
     * 添加一条圆弧路径。
     *
     * @param x               圆孤圆心的 X 坐标
     * @param y               圆弧圆心的 Y 坐标
     * @param radius          圆弧半径
     * @param startAngle      开始角度
     * @param endAngle        结束角度
     * @param anticlockwise   是否逆时针绘制( 1 是， 0 不是)
     */
    arc(
        x: number,
        y: number,
        radius: number,
        startAngle: number,
        endAngle: number,
        anticlockwise: number
    ): void {
        this.addCommand(
            "Arc",
            x,
            y,
            radius,
            startAngle,
            endAngle,
            anticlockwise
        );
    }

    /**
     * 根据控制点和半径添加一条圆弧路径，使用直线连接前一个点。
     *
     * @param x1        点 1 的 x 坐标
     * @param y1        点 1 的 y 坐标
     * @param x2        点 2 的 x 坐标
     * @param y2        点 1 的 y 坐标
     * @param radius    半径
     */
    arcTo(
        x1: number,
        y1: number,
        x2: number,
        y2: number,
        radius: number
    ): void {
        this.addCommand("ArcTo", x1, y1, x2, y2, radius);
    }

    /**
     * 添加二次贝塞尔曲线
     *
     * @param cp1x    控制点 X 坐标
     * @param cp1y    控制点 Y 坐标
     * @param x       结束点 X 坐标
     * @param y       结束点 Y 坐标
     */
    quadraticCurveTo(cp1x: number, cp1y: number, x: number, y: number): void {
        this.addCommand("Q", cp1x, cp1y, x, y);
    }

    /**
     * 添加三次贝塞尔曲线
     *
     * @param cp1x    起始点控制点 X 坐标
     * @param cp1y    起始点控制点 Y 坐标
     * @param cp2x    结束点控制点 X 坐标
     * @param cp2y    结束点控制点 Y 坐标
     * @param x       结束点 X 坐标
     * @param y       结束点 Y 坐标
     */
    bezierCurveTo(
        cp1x: number,
        cp1y: number,
        cp2x: number,
        cp2y: number,
        x: number,
        y: number
    ): void {
        this.addCommand("C", cp1x, cp1y, cp2x, cp2y, x, y);
    }

    /**
     * 添加椭圆
     *
     * @param cx          圆点 X 坐标
     * @param cy          圆点 Y 坐标
     * @param rx          水平半径
     * @param ry          垂直半径
     */
    ellipse(cx: number, cy: number, rx: number, ry: number): void {
        this.addCommand("Ellipse", cx, cy, rx, ry);
    }

    /**
     * 添加矩形
     *
     * @param x           X 坐标
     * @param y           Y 坐标
     * @param w           宽度
     * @param h           高度
     */
    rect(x: number, y: number, w: number, h: number): void {
        this.addCommand("M", x, y);
        this.addCommand("L", x + w, y);
        this.addCommand("L", x + w, y + h);
        this.addCommand("L", x, y + h);
        this.addCommand("Z");
    }

    /**
     * 添加多边形
     *
     * @param pList   点集数组
     */
    polygon(pList: SPoint[]): void {
        if (pList.length > 2) {
            this.addCommand("M", pList[0].x, pList[0].y);
            for (let i = 1; i < pList.length; i++) {
                this.addCommand("L", pList[i].x, pList[i].y);
            }
            this.addCommand("Z");
        }
    }
}
