/*
 * *********************************************************************************************************************
 *
 *          !!
 *        .F88X
 *        X8888Y
 *      .}888888N;
 *        i888888N;        .:!              .I$WI:
 *          R888888I      .'N88~            i8}+8Y&8"l8i$8>8W~'>W8}8]KW+8IIN"8&
 *          .R888888I    .;N8888~          .X8'  "8I.!,/8"  !%NY8`"8I8~~8>,88I
 *            +888888N;  .8888888Y                                  "&&8Y.}8,
 *            ./888888N;  .R888888Y        .'}~    .>}'.`+>  i}!    "i'  +/'  .'i~  !11,.:">,  .~]!  .i}i
 *              ~888888%:  .I888888l      .]88~`1/iY88Ii+1'.R$8$8]"888888888>  Y8$  W8E  X8E  W8888'188Il}Y88$*
 *              18888888    E8888881    .]W%8$`R8X'&8%++N8i,8N%N8+l8%`  .}8N:.R$RE%N88N%N$K$R  188,FE$8%~Y88I
 *            .E888888I  .i8888888'      .:$8I;88+`E8R:/8N,.>881.`$8E/1/]N8X.Y8N`"KF&&FK!'88*."88K./$88%RN888+~
 *            8888888I  .,N888888~        ~88i"8W,!N8*.I88.}888%F,i$88"F88"  888:E8X.>88!i88>`888*.}Fl1]*}1YKi'
 *          i888888N'      I888Y          ]88;/EX*IFKFK88X  K8R  .l8W  88Y  ~88}'88E&%8W.X8N``]88!.$8K  .:W8I
 *        .i888888N;        I8Y          .&8$  .X88!  i881.:%888>I88  ;88]  +88+.';;;;:.Y88X  18N.,88l  .+88/
 *      .:R888888I
 *      .&888888I                                          Copyright (c) 2009-2020.  博锐尚格科技股份有限公司
 *        ~8888'
 *        .!88~                                                                     All rights reserved.
 *
 * *********************************************************************************************************************
 */

import { SFont, SLine, SPaintEngineType, SPath, SPoint, SRect } from "..";
import { SPaintState } from "./SPaintState";
import { SMatrix } from "@persagy-web/base";

/**
 * 绘制引擎基类
 *
 * @author 庞利祥 <sybotan@126.com>
 */
export abstract class SPaintEngine {
    /** 当关状态 */
    private _state = new SPaintState();
    get state(): SPaintState {
        return this._state;
    }
    set state(value: SPaintState) {
        this._state = value;
    }

    /** 状态栈 */
    private stateStack = new Array<SPaintState>();

    /** 绘制引擎类型 */
    abstract get type(): SPaintEngineType;

    /**
     * 保存 painter 状态
     */
    save(): void {
        let state = new SPaintState(this.state);
        this.stateStack.push(state);
    }

    /**
     * 恢复 painter 状态
     */
    restore(): void {
        if (this.stateStack.length > 0) {
            this.state = this.stateStack.pop() as SPaintState;
        }
    }

    // =================================================================================================================
    // 变换相关
    /**
     * 平移变换
     *
     * @param x     X 轴方向平移
     * @param y     Y 辆方向平移
     */
    translate(x: number, y: number): void {
        this.state.matrix.translate(x, y);
    }

    /**
     * 缩放
     *
     * @param x     X 轴方向缩放
     * @param y     Y 轴方向缩放
     */
    scale(x: number, y: number): void {
        this.state.matrix.scale(x, y);
    }

    /**
     * 旋转
     *
     * @param angle     旋转角度(单位弧度)
     */
    rotate(angle: number): void {
        this.state.matrix.rotate(0, 0, angle);
    }

    /**
     * 将当前的变形矩阵乘上一个基于自身参数的矩阵
     *
     * @param m11   水平方向的缩放
     * @param m12   水平方向的倾斜偏移
     * @param m21   竖直方向的倾斜偏移
     * @param m22   竖直方向的缩放
     * @param dx    水平方向的移动
     * @param dy    竖直方向的移动
     */
    transform(
        m11: number,
        m12: number,
        m21: number,
        m22: number,
        dx: number,
        dy: number
    ): void {}

    /**
     * 将当前的变形矩阵重置为单位矩阵，再将当前的变形矩阵乘上一个基于自身参数的矩阵
     *
     * @param m11   水平方向的缩放
     * @param m12   水平方向的倾斜偏移
     * @param m21   竖直方向的倾斜偏移
     * @param m22   竖直方向的缩放
     * @param dx    水平方向的移动
     * @param dy    竖直方向的移动
     */
    setTransform(
        m11: number,
        m12: number,
        m21: number,
        m22: number,
        dx: number,
        dy: number
    ): void {}

    /**
     * 重置当前变形为单位矩阵。等价于调用 setTransform(1, 0, 0, 1, 0, 0)
     */
    resetTransform(): void {
        this.state.matrix = new SMatrix();
    }

    // =================================================================================================================
    // 绘制图形

    /**
     * 清空矩形区域
     *
     * @param rect  矩形
     */
    abstract clearRect(rect: SRect): void;

    /**
     * 绘制空心矩形
     *
     * @param rect  矩形
     */
    abstract drawRect(rect: SRect): void;

    /**
     * 绘制圆形
     *
     * @param cx    圆心 X 坐标
     * @param cy    圆心 Y 坐标
     * @param r     圆半径
     */
    abstract drawCircle(cx: number, cy: number, r: number): void;

    /**
     * 绘制椭圆
     *
     * @param cx    圆点 X 坐标
     * @param cy    圆点 Y 坐标
     * @param rx    水平半径
     * @param ry    垂直半径
     */
    abstract drawEllipse(cx: number, cy: number, rx: number, ry: number): void;

    /**
     * 绘制线段
     *
     * @param line  线段
     */
    abstract drawLine(line: SLine): void;

    /**
     * 绘制折线
     *
     * @param points    折线折点
     */
    abstract drawPolyline(points: SPoint[]): void;

    /**
     * 绘制多边形
     *
     * @param points    多边形顶点
     */
    abstract drawPolygon(points: SPoint[]): void;

    /**
     * 绘制路径
     *
     * @param path  路径
     */
    abstract drawPath(path: SPath): void;

    /**
     * 绘制文本
     *
     * @param text          文本内容
     * @param x             X 坐标
     * @param y             Y 坐标
     * @param maxWidth      最大宽度
     */
    abstract drawText(
        text: string,
        x: number,
        y: number,
        maxWidth?: number
    ): void;

    /**
     * 绘制图片
     *
     * @param img       图片
     * @param x         X 坐标
     * @param y         Y 坐标
     * @param width     宽度
     * @param height    高度
     */
    abstract drawImage(
        img: CanvasImageSource,
        x: number,
        y: number,
        width?: number,
        height?: number
    ): void;

    /**
     * 预测量文本宽度
     *
     * @param text  预测的文本
     */
    abstract textWidth(text: string): number;

    /**
     * 修改字体
     *
     * @param font  字体
     */
    abstract changeFont(font: SFont): void;

    /**
     * 绘制带导角空心矩形
     *
     * @param rect    矩形
     * @param r       导角半径
     */
    abstract drawRoundRect(rect: SRect, r: number): void;
}
