/*
 * *********************************************************************************************************************
 *
 *          !!
 *        .F88X
 *        X8888Y
 *      .}888888N;
 *        i888888N;        .:!              .I$WI:
 *          R888888I      .'N88~            i8}+8Y&8"l8i$8>8W~'>W8}8]KW+8IIN"8&
 *          .R888888I    .;N8888~          .X8'  "8I.!,/8"  !%NY8`"8I8~~8>,88I
 *            +888888N;  .8888888Y                                  "&&8Y.}8,
 *            ./888888N;  .R888888Y        .'}~    .>}'.`+>  i}!    "i'  +/'  .'i~  !11,.:">,  .~]!  .i}i
 *              ~888888%:  .I888888l      .]88~`1/iY88Ii+1'.R$8$8]"888888888>  Y8$  W8E  X8E  W8888'188Il}Y88$*
 *              18888888    E8888881    .]W%8$`R8X'&8%++N8i,8N%N8+l8%`  .}8N:.R$RE%N88N%N$K$R  188,FE$8%~Y88I
 *            .E888888I  .i8888888'      .:$8I;88+`E8R:/8N,.>881.`$8E/1/]N8X.Y8N`"KF&&FK!'88*."88K./$88%RN888+~
 *            8888888I  .,N888888~        ~88i"8W,!N8*.I88.}888%F,i$88"F88"  888:E8X.>88!i88>`888*.}Fl1]*}1YKi'
 *          i888888N'      I888Y          ]88;/EX*IFKFK88X  K8R  .l8W  88Y  ~88}'88E&%8W.X8N``]88!.$8K  .:W8I
 *        .i888888N;        I8Y          .&8$  .X88!  i881.:%888>I88  ;88]  +88+.';;;;:.Y88X  18N.,88l  .+88/
 *      .:R888888I
 *      .&888888I                                          Copyright (c) 2009-2020.  博锐尚格科技股份有限公司
 *        ~8888'
 *        .!88~                                                                     All rights reserved.
 *
 * *********************************************************************************************************************
 */
import { SLine, SPoint, SRect } from "@persagy-web/draw";

/**
 * 数学计算工具类
 *
 * @author 郝建龙 <haojianlong@sagacloud.cn>
 */
export class SMathUtil {
    /**
     * 计算点到点距离
     *
     * @param x1    点 1 的 x 坐标
     * @param y1    点 1 的 y 坐标
     * @param x2    点 2 的 x 坐标
     * @param y2    点 2 的 y 坐标
     * @return 距离
     */
    static pointDistance(
        x1: number,
        y1: number,
        x2: number,
        y2: number
    ): number {
        return Math.sqrt((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2));
    }

    /**
     * 计算矩形是否有交集(外包矩形算法)
     *
     * @param rect1   矩形 1
     * @param rect2   矩形 2
     * @return 矩形是否有交集
     */
    static rectIntersection(rect1: SRect, rect2: SRect): boolean {
        let minX = rect1.x < rect2.x ? rect1.x : rect2.x;
        let minY = rect1.y < rect2.y ? rect1.y : rect2.y;
        let maxX = rect1.right > rect2.right ? rect1.right : rect2.right;
        let maxY = rect1.bottom > rect2.bottom ? rect2.bottom : rect2.bottom;
        return (
            rect1.width + rect2.width > maxX - minX &&
            rect1.height + rect2.height > maxY - minY
        );
    }

    /**
     * 计算线段交点
     *
     * @param line1   线段 1
     * @param line2   线段 2
     * @return 交点 null 平行但不重合 'repeat' 重合
     */
    static lineIntersection(
        line1: SLine,
        line2: SLine
    ): SPoint | null | string {
        let k1 = line1.dy / line1.dx;
        let b1 = line1.y1 - k1 * line1.x1;
        let k2 = line2.dy / line2.dx;
        let b2 = line2.y1 - k2 * line2.x1;
        if (k1 == k2) {
            if (b1 == b2) {
                return "repeat";
            }
            return null;
        }
        let intersectionX = (b2 - b1) / (k1 - k2);
        let intersectionY = k1 * intersectionX + b1;
        let minX = Math.min(line1.x1, line1.x2);
        let maxX = Math.min(line2.x1, line2.x2);
        if (intersectionX >= minX && intersectionX <= maxX) {
            return new SPoint(intersectionX, intersectionY);
        }
        return null;
    }

    /**
     * 转化数据格式为[][]
     *
     * @param  sp  SPoint 类型点数组
     * @return 二维数组类型点数组
     */
    static transferToArray(sp: SPoint[]): number[][] {
        return sp.map((t): number[] => {
            return [t.x, t.y];
        });
    }

    /**
     * 转化数据格式为 SPoint[]
     *
     * @param arr   二维数组类型点数组
     * @return SPoint 类型点数组
     */
    static transferToSPoint(arr: number[][]): SPoint[] {
        return arr.map(
            (t): SPoint => {
                return new SPoint(t[0], t[1]);
            }
        );
    }

    /**
     * 计算轮廓线面积
     *
     * @param arr   轮廓线数组
     * @return 面积
     */
    static calculateArea(arr: SPoint[]): number {
        let sum = 0;
        let n = arr.length;
        arr[n] = arr[0];
        for (let i = 1; i <= n; i++) {
            sum += arr[i].x * arr[i - 1].y - arr[i - 1].x * arr[i].y;
        }
        return sum / 2;
    }
}
